/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "fSeries.hh"
#include "DVecType.hh"
#include <stdexcept>
#include <iostream>

using namespace std;
using namespace containers;

//======================================  Constructor
fSeries::fSeries(void)
    : mF0(0), mDf(0), mT0(0), mDt(0), mDelta(0), mDSMode(kEmpty)
{}

//======================================  Constructor
fSeries::fSeries(const fSeries& x)
{
    *this = x;
}

//======================================  Constructor
fSeries::fSeries(freq_type f0, freq_type dF, const Time& t0, Interval dT, 
		 const DVector& dV) 
    : mF0(f0), mDf(dF), mT0(t0), mDt(dT), mDelta(0), mData(dV.clone())
{
    setDSMode();
}

//======================================  Constructor
fSeries::fSeries(freq_type f0, freq_type dF, const Time& t0, Interval dT, 
		 DVector* dV)
    : mF0(f0), mDf(dF), mT0(t0), mDt(dT), mDelta(0), mData(dV)
{
    setDSMode();
}

//======================================  Clone an fSeries
void
fSeries::appName(const string& txt) {
    mName += txt;
}

//======================================  Clone an fSeries
fSeries* 
fSeries::clone(void) const {
    return new fSeries(*this);
}

//======================================  Extend an fSeries
void 
fSeries::extend(freq_type fmax) {
    if (mDf <= 0 || null()) 
	throw runtime_error("fSeries: Attempted to extend empty series");
    size_type nNew = size_type((fmax - mF0)/mDf + 0.5) + 1;
    mData->Extend(nNew);
}

//======================================  Extract a subseries.
fSeries 
fSeries::extract(freq_type f0, freq_type dF) const {
    freq_type fmax = f0 + dF;
    if (f0 < mF0) f0 = mF0;
    size_type inx0 = getBin(f0);
    size_type inx1 = getBin(fmax) + 1;
    if (empty() || inx1 <= inx0) return fSeries(f0, mDf, mT0, mDt);
    return fSeries(getBinF(inx0), mDf, mT0, mDt, mData->Extract(inx0, inx1-inx0));
}

//======================================  get FSType
fSeries::FSType
fSeries::getFSType(void) const {
    return kSeries;
}

//======================================  Resample series
fSeries
fSeries::interpolate(freq_type fmin, freq_type fmax, freq_type df, 
		     bool l) const {

    //----------------------------------  Set the return series.
    fSeries r(fmin, df, mT0, mDt);
    if (empty() || fmax < fmin) return r;

    //----------------------------------  See if extract will work...
    size_type inc = size_type(df / mDf + 0.5);
    if (double(inc)*mDf == df && mF0 == 0) {
 	size_type iMax = getBin(fmax) + inc;
        if (inc == 1) r.mData.reset(mData->Extract(0, iMax));
	else          r.mData.reset(mData->Extract(0, iMax/inc, inc));
	size_type iMin = r.getBin(fmin);
	if (iMin) r.mData->sub(0, *r.mData, 0, iMin);
	r.mDSMode = kFolded;

    //----------------------------------  Do the full interpolation.
    } else {
	size_type nBin = size_type((fmax - fmin)/df + 1e-4) + 1;
        DVecType<dComplex> dv(nBin);

	//------------------------------  Get the input series packing
	size_type nF = getNStep();
	freq_type fHi = getHighFreq();

	//------------------------------  Interpolate (slowly but surely).
	if (refDVect().C_data()) {
	    const fComplex* p = reinterpret_cast<const fComplex*>(mData->refData());
	    for (size_type i=0; i<nBin; ++i) {
		freq_type f = fmin + df*double(i);
		if (f < mF0 || f>fHi ) {
		    dv[i] = dComplex(0);
		} else {
		    size_type inx = size_type((f - mF0)/mDf);
		    freq_type f0  = getBinF(inx);
		    if (inx == nF) {
			inx--;
			f0 -= mDf;
		    }
		    double a   = (f - f0) / mDf;
		    if (l) dv[i] = exp(log(p[inx])*(1-a) + log(p[inx+1])*a);
		    else   dv[i] = p[inx]*(1.0-a) + p[inx+1]*a;
		}
	    }
	} else if (refDVect().W_data()) {
	    const dComplex* p = reinterpret_cast<const dComplex*>(mData->refData());
	    for (size_type i=0; i<nBin; ++i) {
		freq_type f = fmin + df*double(i);
		if (f < mF0 || f>fHi ) {
		    dv[i] = dComplex(0);
		} else {
		    size_type inx = size_type((f - mF0)/mDf);
		    freq_type f0  = getBinF(inx);
		    if (inx == nF) {
			inx--;
			f0 -= mDf;
		    }
		    double    a   = (f - f0) / mDf;
		    if (l) dv[i] = exp(log(p[inx])*(1-a) + log(p[inx+1])*a);
		    else   dv[i] = p[inx]*(1.0-a) + p[inx+1]*a;
		}
	    }
	} else if (refDVect().D_data()) {
	    const double* p = reinterpret_cast<const double*>(mData->refData());
	    for (size_type i=0; i<nBin; ++i) {
		freq_type f = fmin + df*double(i);
		if (f < mF0 || f>fHi ) {
		    dv[i] = dComplex(0);
		} else {
		    size_type inx = size_type((f - mF0)/mDf);
		    freq_type f0  = getBinF(inx);
		    if (inx == nF) {
			inx--;
			f0 -= mDf;
		    }
		    double    a   = (f - f0) / mDf;
		    if (l) dv[i] = exp(log(p[inx])*(1-a) + log(p[inx+1])*a);
		    else   dv[i] = p[inx]*(1.0-a) + p[inx+1]*a;
		}
	    }
	} else {
	    double* p = new double[nF];
	    getData(nF, p);
	    for (size_type i=0; i<nBin; ++i) {
		freq_type f = fmin + df*double(i);
		if (f < mF0 || f>fHi ) {
		    dv[i] = dComplex(0);
		} else {
		    size_type inx = size_type((f - mF0)/mDf);
		    freq_type f0  = getBinF(inx);
		    if (inx == nF) {
			inx--;
			f0 -= mDf;
		    }
		    double a   = (f - f0) / mDf;
		    if (l) dv[i] = exp(log(p[inx])*(1-a) + log(p[inx+1])*a);
		    else   dv[i] = p[inx]*(1.0-a) + p[inx+1]*a;
		}
	    }
	    delete[] p;
	}
	r.setData(fmin, df, dv);
    }
    return r;
}

//======================================  Print out the fSeries contents 
std::ostream&
fSeries::Dump(std::ostream& out) const {
    out << "fSeries " << mName << ": Start time " << getStartTime() 
	<< " End Time " << getEndTime() << endl;
    out << "Series type: ";
    switch(getFSType()) {
    case kSeries:
      out << "fSeries";
      break;
    case kDFT:
      out << "DFT";
      break;
    case kPSD:
      out << "PSD";
      break;
    case kCSD:
      out << "CSD";
      break;
    case kASD:
      out << "ASD";
      break;
    }
    out << "  Storage type: ";
    switch(mDSMode) {
    case kEmpty:
      out << "None" << endl;
      break;
    case kFolded:
      out << "Folded" << endl;
      break;
    case kFull:
      out << "Full" << endl;
      break;
    }
    out << "Low Frequency " << getLowFreq() << " High Frequency " 
	<< getHighFreq() << " No. frequency steps " << getNStep() << endl;
    out << "Time Series start: " << mT0 << " Interval: " << mDt
	<< " Step: " << mDelta << endl;
    if (null()) return out;
    else        return mData->Dump(out);
}

//======================================  Erase all data from series
void 
fSeries::clear(void) {
    if (!null()) mData->Clear();
    mDSMode = kEmpty;
}

//======================================  Remove negative frequencies
void
fSeries::fold(void) {
    if (empty() || single_sided()) return;
    size_type     N = size();
    size_type ibin0 = N/2;
    mF0 = getBinF(ibin0);
    mData.reset( mData->Extract(ibin0, N-ibin0) );
    mDSMode = kFolded;
}

//======================================  Constructor
void 
fSeries::setData(freq_type f0, freq_type df, const DVector& data) {
    setData(f0, df, data.clone());
}

//======================================  setData
void 
fSeries::setData(freq_type f0, freq_type df, DVector* data) {
    mF0   = f0;
    mDf   = df;
    mData.reset(data);  // auto_ptr is protected from resetting the same pointer
    setDSMode();
}

//======================================  Set the data storage mode
void
fSeries::setDSMode(void) {
    if (mDf <= 0.0) throw runtime_error("fSeries: Invalid frequency step.");
    if (!size())      mDSMode = kEmpty;
    else if (mF0 < 0) mDSMode = kFull;
    else              mDSMode = kFolded;
}

//======================================  Set the name string
void 
fSeries::setName(const std::string& name) {
    mName = name;
}

//======================================  Set the source sample time
void 
fSeries::setSampleTime(Interval dT) {
    mDelta = dT;
}

//======================================  Set the time span
void 
fSeries::setTimeSpan(const Time& t0, Interval dT) {
    mT0 = t0;
    mDt = dT;
}

//======================================  Reserve space in the data vector
void 
fSeries::reserve(size_type len) {
    if (!null()) refDVect().reserve(len);
}

//======================================  Assignment operators.
fSeries& 
fSeries::operator =(const fSeries& x) {
    mName   = x.mName;
    mF0     = x.mF0;
    mDf     = x.mDf;
    mT0     = x.mT0;
    mDt     = x.mDt;
    mDSMode = x.mDSMode;
    mDelta  = x.mDelta;
    if (!empty()) mData.reset(x.refDVect().clone());
    else          mData.reset(0);
    return *this;
}

//======================================  Bias operator
fSeries& 
fSeries::operator +=(double bias) {
    if (!empty()) refDVect() += bias;
    return *this;
}

//======================================  Addition operator
fSeries& 
fSeries::operator +=(const fSeries& rhs) {
    if (empty() || size() != rhs.size()) 
	throw runtime_error("fSeries: Add an unequal length fSeries");
    refDVect() += rhs.refDVect();
    return *this;
}

//======================================  Subtract operator
fSeries& 
fSeries::operator -=(const fSeries& rhs) {
    if (empty() || size() != rhs.size()) 
	throw runtime_error("fSeries: Subtract an unequal length fSeries");
    refDVect() -= rhs.refDVect();
    return *this;
}

//======================================  Scale operator
fSeries& 
fSeries::operator *=(double scale) {
    if (!empty()) refDVect() *= scale;
    return *this;
}

//======================================  Multiply operator
fSeries& 
fSeries::operator *=(const fSeries& fs) {
    if (empty() || size() != fs.size()) 
	throw runtime_error("fSeries: Multiply by an unequal length fSeries");
    refDVect() *= fs.refDVect();
    return *this;
}

//======================================  Divide operator
fSeries& 
fSeries::operator /=(const fSeries& fs) {
    if (empty() || size() != fs.size()) 
	throw runtime_error("fSeries: Divide by an unequal length series");
    refDVect() /= fs.refDVect();
    return *this;
}

//======================================  Parenthesis operator.
dComplex 
fSeries::operator()(freq_type freq) const {
    return refDVect().getCplx(getBin(freq));
}

//======================================  Set data template.
template<class T>
void 
fSeries::setData(freq_type f0, freq_type df, size_type len, const T* data) {
    mF0   = f0;
    mDf   = df;
    mData.reset( new DVecType<T>(len, data) );
    setDSMode();
}

//======================================  Template instanciation
#if !defined (__SUNPRO_CC)
template
void
fSeries::setData<float>(freq_type f0, freq_type df, size_type len, 
			const float* data);

template
void
fSeries::setData<double>(freq_type f0, freq_type df, size_type len, 
			 const double* data);

template
void
fSeries::setData<fComplex>(freq_type f0, freq_type df, size_type len, 
			   const fComplex* data);

template
void
fSeries::setData<dComplex>(freq_type f0, freq_type df, size_type len, 
			   const dComplex* data);
#endif

//======================================  Type template Constructor
template<class T>
fSeries::fSeries(freq_type f0, freq_type dF, const Time& t0, Interval dT, 
		 size_type NData, const T* Data) 
    : mF0(f0), mDf(dF), mT0(t0), mDt(dT), mDelta(0)
{
    if (NData) mData.reset(new DVecType<T>(NData, Data));
    setDSMode();
}

#if !defined (__SUNPRO_CC)
template
fSeries::fSeries(freq_type f0, freq_type dF, const Time& t0, 
		 Interval dT, size_type len, const float* data);

template
fSeries::fSeries(freq_type f0, freq_type dF, const Time& t0, 
		 Interval dT, size_type len, const double* data);

template
fSeries::fSeries(freq_type f0, freq_type dF, const Time& t0, 
		 Interval dT, size_type len, const fComplex* data);

template
fSeries::fSeries(freq_type f0, freq_type dF, const Time& t0, 
		 Interval dT, size_type len, const dComplex* data);
#endif
