/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "NDS2Socket.hh"
#include "FrWriter.hh"
#include "TSeries.hh"
#include <vector>
#include <stdexcept>
#include <string>
#include <iostream>

//========================================  NDS2Frame class
class NDS2Frame {
public:
    typedef unsigned long gps_type;
public:
    NDS2Frame(int argc, const char* argv[]);
    ~NDS2Frame(void);
    int main_loop(void);
    void syntax(void) const;
private:
    typedef std::pair<std::string, TSeries*> chan_node;
    typedef std::vector<chan_node> chan_list;
    typedef chan_list::const_iterator const_chan_iter;
    std::string mServer;
    int         mPort;
    gps_type    mStartGPS;
    gps_type    mEndGPS;
    gps_type    mStride;
    std::string mFrameFile;
    int         mDebug;
    bool        mOnline;
    chan_list   mChanList;
    NDS2Socket  mNDS;
    FrWriter    mWriter;
};

using namespace std;

//========================================  Main function
int 
main(int argc, const char* argv[]) {
    try {
	NDS2Frame nds2fr(argc, argv);
	nds2fr.main_loop();
    } catch (std::exception& e) {
	cerr << "Caught exception: " << e.what() << endl;
    }
}

//========================================  Print command line syntax
void
NDS2Frame::syntax(void) const {
    cout << "NDS2Frame command line syntax:" << endl;
    cout << "NDS2Frame [--server <server-ip>] [--port <port>] [--online] " 
	 << "[--help] " << endl;
    cout << "          [--start-gps <start-gps>] [--end-gps <end-gps>] "
	 << "[--stride <stride>] " << endl;
    cout << "          [--out-frame <file-name>] [--debug <debug-lvl>] "
	 << "<channel-1> [... <channel-n>]" << endl;
    cout << endl;
    cout << "<server-ip>  Server address:port [localhost:31200]" << endl;
    cout << "<port>       Server port [0]" << endl;
    cout << "--online     Get online data" << endl;
    cout << "--help       Print this message." << endl;
    cout << "<start-gps>  Start GPS tiem (ignored in online requests" 
	 << endl;
    cout << "<end-gps>    End GPS time (number of data seconds online)"
	 << endl;
    cout << "<stride>     Number of seconds per request (frame length)" 
	 << endl;
    cout << "<file-name>  Output file name template [NDS2Frame-%s-%1n.gwf]" 
	 << endl;
    cout << "<debug-lvl>  Debug print level" << endl;
    cout << "<channel-1> ... <channel-n> list of channels to be written to frame"
	 << endl;
    cout << endl;
}

//========================================  NDS2Frame constructor
NDS2Frame::NDS2Frame(int argc, const char* argv[])
    : mServer("localhost:31200"), mPort(0), mStartGPS(0), mEndGPS(0), 
      mStride(1), mFrameFile("NDS2Frame-%s-%1n.gwf"), mDebug(0), 
      mOnline(false), mWriter("LIGO_NDS", 0)
{

    //------------------------------------  Parse some arguments
    for (int i=1; i<argc; ++i) {
	string argi = argv[i];
	if (argi[0] != '-') {
	    mChanList.push_back(chan_node(argi, new TSeries));
	} else if (argi == "--debug") {
	    mDebug = strtol(argv[++i], 0, 0);
	} else if (argi == "-e" || argi == "--end-gps") {
	    mEndGPS = strtoul(argv[++i], 0, 0);
	} else if (argi == "--help") {
	    syntax();
	    throw runtime_error("Help request");
	} else if (argi == "--online") {
	    mOnline = true;
	} else if (argi == "-o" || argi == "--out-frame") {
	    mFrameFile = argv[++i];
	} else if (argi == "-p" || argi == "--port") {
	    mPort = strtol(argv[++i], 0, 0);
	} else if (argi == "--server") {
	    mServer = argv[++i];
	} else if (argi == "-s" || argi == "--start-gps") {
	    mStartGPS = strtoul(argv[++i], 0, 0);
	} else if (argi == "--stride") {
	    mStride = strtoul(argv[++i], 0, 0);
	} else {
	    syntax();
	    throw runtime_error(string("Unknown command argument: ")+argi);
	}
    }

    //------------------------------------ Open the server client
    string::size_type inx = mServer.find(":");
    if (inx == string::npos) {
	if (!mPort) mPort = 31200;
    } else if (!mPort) {
	mPort = strtol(mServer.c_str() + inx + 1, 0, 0);
	mServer.erase(inx);
    } else {
	mServer.erase(inx);
    }

    mNDS.setDebug(mDebug);
    if (mDebug) cerr << "Connecting socket to server: " << mServer 
		     << ":" << mPort << endl;
    int rc = mNDS.open(mServer, mPort);
    if (rc) {
	throw runtime_error(string("Unable to connect client to: ")+mServer);
    }

    chantype ctype = cUnknown;
    if (mOnline) {
	ctype = cOnline;
	if (mEndGPS == 0) mEndGPS = 86400;
    }

    int N = mChanList.size();
    for (int i=0; i<N; ++i) {
	mNDS.AddChannel(mChanList[i].first, ctype, 0);
	mWriter.addChannel(mChanList[i].first.c_str(), &mChanList[i].second);
    }
    mWriter.setDebug(mDebug);
}

//========================================  NDS2Frame constructor
int
NDS2Frame::main_loop(void) {
    int rc;
    int nChan = mChanList.size();

    //------------------------------------  Request online data
    if (mOnline) {
	rc = mNDS.RequestOnlineData(mStride);
	if (rc) throw runtime_error("RequestData failed.");
    }

    //------------------------------------  Loop over strides
    for (gps_type t0=mStartGPS; t0<mEndGPS; t0 += mStride) {

	//--------------------------------  Request a bunch of data.
	int rc = 0;
	if (!mOnline) {
	    rc = mNDS.RequestData(t0, mStride);
	    if (rc) throw runtime_error("RequestData failed.");
	}

	//--------------------------------  Get the data
	rc = mNDS.GetData();
	if (rc <= 0) throw runtime_error("GetData failed.");

	//--------------------------------  Set frame size parameters
	Time t(mNDS.mRecvBuf.ref_header().GPS);
	Interval delta(mStride);

	//--------------------------------  Loop over channels, unpack TSeries
	union {
	    const char*   c;
	    const double* d;
	    const float*  f;
	    const short*  s;
	    const int*    i;
	} p;

	for (int i=0; i<nChan; ++i) {
	    DAQC_api::const_channel_iter 
		cp = mNDS.FindChannel(mChanList[i].first);
	    Interval dt(1./cp->mRate);
	    unsigned long nw = cp->nwords(mStride);
	    p.c = mNDS.mRecvBuf.ref_data() + cp->mBOffset;
	    if (mDebug) cerr << "Get channel: " << cp->mName << " data-type: "
			     << cp->mDatatype << endl;
	    switch (cp->mDatatype) {
	    case _16bit_integer:
		mChanList[i].second->setData(t, dt, p.s, nw);
		break;
	    case _32bit_integer:
		mChanList[i].second->setData(t, dt, p.i, nw);
		break;
	    case _32bit_float:
		mChanList[i].second->setData(t, dt, p.f, nw);
		break;
	    case _64bit_double:
		mChanList[i].second->setData(t, dt, p.d, nw);
		break;
	    default:
		throw runtime_error("Unsupported data type");
	    }
	}
	mWriter.buildFrame(t, delta);
	mWriter.addWriterHistory();

	//------------------------------  Generate the file name
	char file[1024];
	Time tBogus = t;
	tBogus.setN(delta);
	TimeStr(tBogus, file, mFrameFile.c_str());

	//------------------------------  Write the frame
	rc = mWriter.open(file);
	if (rc) {
	    cerr << "NDS2Frame unable to open output file: " << file << endl;
	    return rc;
	}
	mWriter.writeFrame();
	mWriter.close();
    }
    return 0;
}

NDS2Frame::~NDS2Frame(void) {
    mWriter.close();
    int N = mChanList.size();
    for (int i=0; i<N; ++i) {
	delete mChanList[i].second;
	mChanList[i].second = 0;
    }
}
